########## R function: NorRSinferMFVB ###########

# For performing approximate inference for the
# parameters in a Bayesian Normal random sample
# model via mean field variational Bayes.

# Last changed: 21 JUL 2015

NorRSinferMFVB <- function(x,mu.mu=0,sigsq.mu=1e5,A=1e5,maxIter=100,
                         tolerance=1e-10,epsilon=0,plotML=FALSE,
                         convergeType="MaximumIterations")
{
   # Create function for approximate marginal log-likelihood:

   logML <- function(n,sigsq.mu,A,mu.q.mu,sigsq.q.mu,mu.q.recip.sigsq,
                     mu.q.recip.a)
   {
      ans <- 0.5 - 0.5*n*log(2*pi) - 2*log(pi) + lgamma((n+1)/2)
      ans <- ans - 0.5*log(sigsq.mu) - log(A)
      ans <- ans - 0.5*((mu.q.mu^2) + sigsq.q.mu)/sigsq.mu
      ans <- ans + 0.5*log(sigsq.q.mu) - 0.5*(n+1)*log((n+1)/(2*mu.q.recip.sigsq))
      ans <- ans - log(mu.q.recip.sigsq + (1/A^2)) + mu.q.recip.sigsq*mu.q.recip.a
      return(ans)
   }
  
   # Extract sample size:

   n <- length(x)

   # Compute summary statistics:

   sumx <- sum(x) ; sumxsq <- sum(x^2)

   # Update the message natural parameters
   # until convergence:

   converged <- FALSE ; itNum <- 0;
   logMLgrid <- rep(NA,maxIter)
   logMLcurr <- -Inf

   mu.q.recip.sigsq <- 1
   mu.q.recip.a <- 1

   while (!converged)
   {
      logMLprev <- logMLcurr
      itNum <- itNum + 1

      # Perform updates for q*(mu) parameters:

      sigsq.q.mu <- 1/(mu.q.recip.sigsq*n + (1/sigsq.mu))
                                     
      mu.q.mu <- mu.q.recip.sigsq*sigsq.q.mu*sum(x)
 
      # Perform updates for q*(a) parameters:

      mu.q.recip.a <- 1/(mu.q.recip.sigsq + (1/A^2))
      
      # Perform updates for q*(sigsq) parameters:

      A.q.sigsq <- 0.5*(n+1)
      B.q.sigsq <- (0.5*(sum((x-mu.q.mu)^2) + n*sigsq.q.mu)
                    + mu.q.recip.a)
      mu.q.recip.sigsq <- A.q.sigsq/B.q.sigsq
      
      if (itNum>=maxIter) 
      {
         converged <- TRUE
         #print("WARNING: maximum number of iterations exceeded.")
      }

      if (convergeType!="MaximumIterations")
      {
         # Evaluate current approximate log-likelihood:

         logMLcurr <- logML(n,sigsq.mu,A,mu.q.mu,sigsq.q.mu,mu.q.recip.sigsq,
                            mu.q.recip.a)

         logMLgrid[itNum] <- logMLcurr

         # Compute relative error:

         relErr <- abs((logMLcurr/logMLprev)-1)

         # Check `converged' conditions:
       
         if (relErr<tolerance) converged <- TRUE
      }    
   }

   logMLgrid <- logMLgrid[1:itNum]
  
   useWaiting <- FALSE
   if (plotML)
   {
      omitVal <- 0
      iterGrid <- 1:length(logMLgrid)

      if (omitVal>0)
      {
         logMLgrid <- logMLgrid[-(1:omitVal)]
         iterGrid <- iterGrid[-(1:omitVal)]
      }
         
      plot(iterGrid,logMLgrid,type="b",bty="l",xlab="iterations",
           ylab="approximate marginal log-likelihood")
      ans <- readline()
      cat("Hit Enter to continue.\n")
   }
   
   # Return parameters of approximate posterior density functions:
   
   return(list(mu.q.mu=mu.q.mu,sigsq.q.mu=sigsq.q.mu,
               A.q.sigsq=A.q.sigsq,B.q.sigsq=B.q.sigsq))
}
   
########## End of NorRSinferMFVB ###########
   
